#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

import testtools

from observabilityclient.prometheus_client import PrometheusMetric
from observabilityclient.utils import metric_utils
from observabilityclient.v1 import cli


class CliTest(testtools.TestCase):
    def setUp(self):
        super().setUp()
        self.client = mock.Mock()
        self.client.query = mock.Mock()

    def test_list(self):
        metric_names = ['name1', 'name2', 'name3']
        expected = (['metric_name'], [['name1'], ['name2'], ['name3']])
        cli_list = cli.List(mock.Mock(), mock.Mock())

        parser = cli_list.get_parser("metric list")
        test_parsed_args_enabled = parser.parse_args([
        ])
        test_parsed_args_disabled = parser.parse_args([
            "--disable-rbac"
        ])

        with mock.patch.object(metric_utils, 'get_client',
                               return_value=self.client), \
                mock.patch.object(self.client.query, 'list',
                                  return_value=metric_names) as m:
            # NOTE: disable-rbac option is NOOP, RBAC is disabled by default
            ret1 = cli_list.take_action(test_parsed_args_enabled)
            m.assert_called_with(disable_rbac=True)

            ret2 = cli_list.take_action(test_parsed_args_disabled)
            m.assert_called_with(disable_rbac=True)

        self.assertEqual(ret1, expected)
        self.assertEqual(ret2, expected)

    def test_show(self):
        metric = {
            'value': [123456, 12],
            'metric': {'label1': 'value1'}
        }
        prom_metric = [PrometheusMetric(metric)]
        expected = ['label1', 'value'], [['value1', 12]]

        cli_show = cli.Show(mock.Mock(), mock.Mock())

        parser = cli_show.get_parser("metric show")
        test_parsed_args_enabled = parser.parse_args([
            "metric_name"
        ])
        test_parsed_args_disabled = parser.parse_args([
            "metric_name",
            "--disable-rbac"
        ])

        with mock.patch.object(metric_utils, 'get_client',
                               return_value=self.client), \
                mock.patch.object(self.client.query, 'show',
                                  return_value=prom_metric) as m:
            # NOTE: disable-rbac option is NOOP, RBAC is disabled by default
            ret1 = cli_show.take_action(test_parsed_args_enabled)
            m.assert_called_with('metric_name', disable_rbac=True)

            ret2 = cli_show.take_action(test_parsed_args_disabled)
            m.assert_called_with('metric_name', disable_rbac=True)

        self.assertEqual(ret1, expected)
        self.assertEqual(ret2, expected)

    def test_query(self):
        query = ("some_query{label!~'not_this_value'} - "
                 "sum(second_metric{label='this'})")

        metric = {
            'value': [123456, 12],
            'metric': {'label1': 'value1'}
        }

        prom_metric = [PrometheusMetric(metric)]
        expected = ['label1', 'value'], [['value1', 12]]

        cli_query = cli.Query(mock.Mock(), mock.Mock())

        parser = cli_query.get_parser("metric query")
        test_parsed_args_enabled = parser.parse_args([
            query
        ])
        test_parsed_args_disabled = parser.parse_args([
            query,
            "--disable-rbac"
        ])

        with mock.patch.object(metric_utils, 'get_client',
                               return_value=self.client), \
                mock.patch.object(self.client.query, 'query',
                                  return_value=prom_metric) as m:

            # NOTE: disable-rbac option is NOOP, RBAC is disabled by default
            ret1 = cli_query.take_action(test_parsed_args_enabled)
            m.assert_called_with(query, disable_rbac=True)

            ret2 = cli_query.take_action(test_parsed_args_disabled)
            m.assert_called_with(query, disable_rbac=True)

        self.assertEqual(ret1, expected)
        self.assertEqual(ret2, expected)

    def test_delete(self):
        match1 = "some_label_name"
        match2 = "some_label_name2"

        cli_delete = cli.Delete(mock.Mock(), mock.Mock())

        parser = cli_delete.get_parser("metric delete")
        test_parsed_args = parser.parse_args([
            match1, match2,
            "--start", "0",
            "--end", "10"
        ])

        with mock.patch.object(metric_utils, 'get_client',
                               return_value=self.client), \
                mock.patch.object(self.client.query, 'delete') as m:

            cli_delete.take_action(test_parsed_args)
            m.assert_called_with([[match1, match2]], "0", "10")

    def test_clean_combstones(self):
        cli_clean_tombstones = cli.CleanTombstones(mock.Mock(), mock.Mock())

        with mock.patch.object(metric_utils, 'get_client',
                               return_value=self.client), \
                mock.patch.object(self.client.query, 'clean_tombstones') as m:

            cli_clean_tombstones.take_action({})
            m.assert_called_once()

    def test_snapshot(self):
        cli_snapshot = cli.Snapshot(mock.Mock(), mock.Mock())
        file_name = 'some_file_name'

        with mock.patch.object(metric_utils, 'get_client',
                               return_value=self.client), \
                mock.patch.object(self.client.query, 'snapshot',
                                  return_value=file_name) as m:

            ret = cli_snapshot.take_action({})
            m.assert_called_once()
        self.assertEqual(ret, (["Snapshot file name"], [[file_name]]))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from observabilityclient.tests.functional import base
import time


class CliTestFunctionalRBACNOOP(base.CliTestCase):
    """Functional tests for cli commands testing RBAC option."""

    def test_list(self):
        for cmdstr in ('metric list', 'metric list --disable-rbac'):
            cmd_output = self.openstack(
                cmdstr,
                parse_output=True,
            )
            name_list = [item.get('metric_name') for item in cmd_output]
            self.assertIn(
                'ceilometer_image_size',
                name_list
            )
            self.assertIn(
                'up',
                name_list
            )

    def test_show(self):
        for cmdstr in ('metric show up', 'metric show up --disable-rbac'):
            cmd_output = self.openstack(
                cmdstr,
                parse_output=True,
            )
            for metric in cmd_output:
                self.assertEqual(
                    "up",
                    metric["__name__"]
                )
                self.assertEqual(
                    "1",
                    metric["value"]
                )

    def test_query(self):
        for cmdstr in ('metric query up', 'metric query up --disable-rbac'):
            cmd_output = self.openstack(
                cmdstr,
                parse_output=True,
            )
            for metric in cmd_output:
                self.assertEqual(
                    "up",
                    metric["__name__"]
                )
                self.assertEqual(
                    "1",
                    metric["value"]
                )


class CliTestFunctionalAdminCommands(base.CliTestCase):
    """Functional tests for cli admin commands."""

    def test_delete(self):
        test_start_time = int(time.time())
        query_before = self.openstack(
            f'metric query prometheus_ready@{test_start_time} --disable-rbac',
            parse_output=True,
        )

        values = [item.get("__name__") for item in query_before]
        # Check, that the metric is present before the deletion
        self.assertIn(
            "prometheus_ready",
            values
        )

        self.openstack(
            'metric delete prometheus_ready --disable-rbac',
            parse_output=False,
        )

        query_after = self.openstack(
            f'metric query prometheus_ready@{test_start_time} --disable-rbac',
            parse_output=True,
        )
        values = [item.get("__name__") for item in query_after]
        # Check, that the metric is not present after the deletion
        self.assertNotIn(
            "prometheus_ready",
            values
        )

    def test_clean_tombstones(self):
        # NOTE(jwysogla) There is not much to check here
        # except for the fact, that the command doesn't
        # raise an exception. Prometheus doesn't send any
        # data back and we don't have a reliable way to query
        # prometheus that this command did something.
        self.openstack('metric clean-tombstones')

    def test_snapshot(self):
        cmd_output = self.openstack(
            'metric snapshot',
            parse_output=True,
        )
        for name in cmd_output:
            self.assertInOutput(
                time.strftime('%Y%m%d'),
                name.get("Snapshot file name")
            )



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""OpenStackClient Plugin interface."""

from osc_lib import utils


DEFAULT_API_VERSION = '1'
API_NAME = 'observabilityclient'
API_VERSION_OPTION = 'os_observabilityclient_api_version'
API_VERSIONS = {
    '1': 'observabilityclient.v1.client.Client',
}


def make_client(instance):
    """Return a client to the ClientManager.

    Called to instantiate the requested client version.  instance has
    any available auth info that may be required to prepare the client.

    :param ClientManager instance: The ClientManager that owns the new client
    """
    observability_client = utils.get_client_class(
        API_NAME,
        instance._api_version[API_NAME],
        API_VERSIONS)

    client = observability_client(session=instance.session,
                                  adapter_options={
                                      'interface': instance.interface,
                                      'region_name': instance.region_name
                                  })
    return client


def build_option_parser(parser):
    """Add global options.

    Called from openstackclient.shell.OpenStackShell.__init__()
    after the builtin parser has been initialized.  This is
    where a plugin can add global options such as an API version setting.

    :param argparse.ArgumentParser parser: The parser object that has been
        initialized by OpenStackShell.
    """
    parser.add_argument(
        '--os-observability-api-version',
        metavar='<observability-api-version>',
        help='Observability Plugin API version, default=' +
             DEFAULT_API_VERSION +
             ' (Env: OS_OSCPLUGIN_API_VERSION)')
    return parser



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import os
from unittest import mock

from keystoneauth1 import adapter
from keystoneauth1 import session
import testtools

from observabilityclient import prometheus_client
from observabilityclient.utils import metric_utils


class GetConfigFileTest(testtools.TestCase):
    def setUp(self):
        super().setUp()

    def test_current_dir(self):
        with mock.patch.object(os.path, 'exists', return_value=True), \
                mock.patch.object(metric_utils, 'open') as m:
            metric_utils.get_config_file()
        m.assert_called_with(metric_utils.CONFIG_FILE_NAME)

    def test_path_order(self):
        expected = [mock.call(metric_utils.CONFIG_FILE_NAME, 'r'),
                    mock.call(f"{os.environ['HOME']}/.config/openstack/"
                              f"{metric_utils.CONFIG_FILE_NAME}"),
                    mock.call(f"/etc/openstack/"
                              f"{metric_utils.CONFIG_FILE_NAME}")]
        with mock.patch.object(os.path, 'exists', return_value=False) as m:
            ret = metric_utils.get_config_file()
        m.call_args_list == expected
        self.assertIsNone(ret)


class GetPrometheusClientTest(testtools.TestCase):
    def setUp(self):
        super().setUp()
        config_data = 'host: "somehost"\nport: "1234"'
        self.config_file = mock.mock_open(read_data=config_data)("name", 'r')

    def test_get_prometheus_client_from_file(self):
        with mock.patch.object(metric_utils, 'get_config_file',
                               return_value=self.config_file), \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "__init__", return_value=None) as m:
            metric_utils.get_prometheus_client()
        m.assert_called_with("somehost:1234", None, "")

    def test_get_prometheus_client_env_override(self):
        with mock.patch.dict(os.environ,
                             {'PROMETHEUS_HOST': 'env_override'}), \
                mock.patch.object(metric_utils, 'get_config_file',
                                  return_value=self.config_file), \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "__init__", return_value=None) as m:
            metric_utils.get_prometheus_client()
        m.assert_called_with("env_override:1234", None, "")

    def test_get_prometheus_client_no_config_file(self):
        patched_env = {'PROMETHEUS_HOST': 'env_override',
                       'PROMETHEUS_PORT': 'env_port'}
        with mock.patch.dict(os.environ, patched_env), \
                mock.patch.object(metric_utils, 'get_config_file',
                                  return_value=None), \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "__init__", return_value=None) as m:
            metric_utils.get_prometheus_client()
        m.assert_called_with("env_override:env_port", None, "")

    def test_get_prometheus_client_prefix_in_env_variable(self):
        patched_env = {'PROMETHEUS_HOST': 'env_override',
                       'PROMETHEUS_PORT': 'env_port',
                       'PROMETHEUS_ROOT_PATH': 'root_path_env'}
        with mock.patch.dict(os.environ, patched_env), \
                mock.patch.object(metric_utils, 'get_config_file',
                                  return_value=None), \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "__init__", return_value=None) as m:
            metric_utils.get_prometheus_client()
        m.assert_called_with("env_override:env_port", None, "root_path_env")

    def test_get_prometheus_client_missing_configuration(self):
        with mock.patch.dict(os.environ, {}), \
                mock.patch.object(metric_utils, 'get_config_file',
                                  return_value=None), \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "__init__", return_value=None):
            self.assertRaises(metric_utils.ConfigurationError,
                              metric_utils.get_prometheus_client)

    def test_get_prometheus_client_from_keystone_http(self):
        prom_endpoint = "http://localhost:1234/prometheus"
        keystone_session = session.Session()
        with mock.patch.dict(os.environ, {}), \
                mock.patch.object(metric_utils, 'get_config_file',
                                  return_value=None), \
                mock.patch.object(adapter.Adapter, 'get_endpoint',
                                  return_value=prom_endpoint), \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "__init__", return_value=None) as init_m, \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "set_ca_cert") as ca_m:
            metric_utils.get_prometheus_client(keystone_session)
        init_m.assert_called_with(
            "localhost:1234", keystone_session, "prometheus"
        )
        ca_m.assert_not_called()

    def test_get_prometheus_client_from_keystone_https(self):
        prom_endpoint = "https://localhost:1234/prometheus"
        keystone_session = session.Session()
        with mock.patch.dict(os.environ, {}), \
                mock.patch.object(metric_utils, 'get_config_file',
                                  return_value=None), \
                mock.patch.object(adapter.Adapter, 'get_endpoint',
                                  return_value=prom_endpoint), \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "__init__", return_value=None) as init_m, \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "set_ca_cert") as ca_m:
            metric_utils.get_prometheus_client(keystone_session)
        init_m.assert_called_with(
            "localhost:1234", keystone_session, "prometheus"
        )
        ca_m.assert_called_with(True)

    def test_get_prometheus_client_from_env_vars_ipv6(self):
        patched_env = {'PROMETHEUS_HOST': '2607:5300:201:2000::654',
                       'PROMETHEUS_PORT': '1234'}
        with mock.patch.dict(os.environ, patched_env), \
                mock.patch.object(metric_utils, 'get_config_file',
                                  return_value=None), \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "__init__", return_value=None) as m:
            metric_utils.get_prometheus_client()
        m.assert_called_with("[2607:5300:201:2000::654]:1234", None, "")

    def test_get_prometheus_client_from_conf_file_ipv6(self):
        config_data = '''
        host: "2607:5300:201:2000::654"
        port: "80"
        '''
        config_file = mock.mock_open(read_data=config_data)("name", 'r')
        with mock.patch.dict(os.environ, {}), \
                mock.patch.object(metric_utils, 'get_config_file',
                                  return_value=config_file), \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "__init__", return_value=None) as init_m:
            metric_utils.get_prometheus_client()
        init_m.assert_called_with(
            "[2607:5300:201:2000::654]:80", None, ""
        )

    def test_get_prometheus_client_from_keystone_ipv6(self):
        prom_endpoint = "http://[2607:5300:201:2000::654]:80/prometheus"
        keystone_session = session.Session()
        with mock.patch.dict(os.environ, {}), \
                mock.patch.object(metric_utils, 'get_config_file',
                                  return_value=None), \
                mock.patch.object(adapter.Adapter, 'get_endpoint',
                                  return_value=prom_endpoint), \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "__init__", return_value=None) as init_m, \
                mock.patch.object(prometheus_client.PrometheusAPIClient,
                                  "set_ca_cert") as ca_m:
            metric_utils.get_prometheus_client(keystone_session)
        init_m.assert_called_with(
            "[2607:5300:201:2000::654]:80", keystone_session, "prometheus"
        )
        ca_m.assert_not_called()


class FormatLabelsTest(testtools.TestCase):
    def setUp(self):
        super().setUp()

    def test_format_labels_with_normal_labels(self):
        input_dict = {"label_key1": "label_value1",
                      "label_key2": "label_value2"}
        expected = "label_key1='label_value1', label_key2='label_value2'"

        ret = metric_utils.format_labels(input_dict)
        self.assertEqual(expected, ret)

    def test_format_labels_with_quoted_labels(self):
        input_dict = {"label_key1": "'label_value1'",
                      "label_key2": "'label_value2'"}
        expected = "label_key1='label_value1', label_key2='label_value2'"

        ret = metric_utils.format_labels(input_dict)
        self.assertEqual(expected, ret)


class Metrics2ColsTest(testtools.TestCase):
    def setUp(self):
        super().setUp()

    def test_metrics2cols(self):
        metric = {
            'value': [
                1234567,
                5
            ],
            'metric': {
                'label1': 'value1',
                'label2': 'value2',
            }
        }
        input_metrics = [prometheus_client.PrometheusMetric(metric)]
        expected = (['label1', 'label2', 'value'], [['value1', 'value2', 5]])

        ret = metric_utils.metrics2cols(input_metrics)
        self.assertEqual(expected, ret)

    def test_metrics2cols_column_ordering(self):
        metric = {
            'value': [
                1234567,
                5
            ],
            'metric': {
                'a_label1': 'value1',
                'b_label2': 'value2',
            }
        }
        input_metrics = [prometheus_client.PrometheusMetric(metric)]
        expected = (['a_label1', 'b_label2', 'value'],
                    [['value1', 'value2', 5]])

        ret = metric_utils.metrics2cols(input_metrics)
        self.assertEqual(expected, ret)

        metric = {
            'value': [
                1234567,
                5
            ],
            'metric': {
                'b_label1': 'value1',
                'a_label2': 'value2',
            }
        }
        input_metrics = [prometheus_client.PrometheusMetric(metric)]
        expected = (['a_label2', 'b_label1', 'value'],
                    [['value2', 'value1', 5]])

        ret = metric_utils.metrics2cols(input_metrics)
        self.assertEqual(expected, ret)

        metric1 = {
            'value': [
                1234567,
                5
            ],
            'metric': {
                'b_label1': 'value1',
                'a_label2': 'value2',
            }
        }
        metric2 = {
            'value': [
                1234567,
                5
            ],
            'metric': {
                'b_label1': 'value1',
                'a_label2': 'value2',
                'd_label3': 'value3',
                'c_label4': 'value4',
            }
        }
        input_metrics = [prometheus_client.PrometheusMetric(metric1),
                         prometheus_client.PrometheusMetric(metric2)]
        expected = (['a_label2', 'b_label1', 'c_label4', 'd_label3', 'value'],
                    [['value2', 'value1', '', '', 5],
                     ['value2', 'value1', 'value4', 'value3', 5]]
                    )

        ret = metric_utils.metrics2cols(input_metrics)
        self.assertEqual(expected, ret)



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

import requests

import testtools

from observabilityclient import prometheus_client as client


class MetricListMatcher(testtools.Matcher):
    def __init__(self, expected):
        self.expected = expected

    def __str__(self):
        return ("Matches Lists of metrics as returned "
                "by prometheus_client.PremetheusAPIClient.query")

    def metric_to_str(self, metric):
        return (f"Labels: {metric.labels}\n"
                f"Timestamp: {metric.timestamp}\n"
                f"Value: {metric.value}")

    def match(self, observed):
        if len(self.expected) != len(observed):
            description = (f"len(expected) != len(observed) because "
                           f"{len(self.expected)} != {len(observed)}")
            return testtools.matchers.Mismatch(description=description)

        for e in self.expected:
            for o in observed:
                if (e.timestamp == o.timestamp and
                        e.value == o.value and
                        e.labels == o.labels):
                    observed.remove(o)
                    break

        if len(observed) != 0:
            description = "Couldn't match the following metrics:\n"
            for o in observed:
                description += self.metric_to_str(o) + "\n\n"
            return testtools.matchers.Mismatch(description=description)
        return None


class PrometheusAPIClientTestBase(testtools.TestCase):
    def setUp(self):
        super().setUp()

    class GoodResponse:
        def __init__(self):
            self.status_code = 200

        def json(self):
            return {"status": "success"}

    class BadResponse:
        def __init__(self):
            self.status_code = 500

        def json(self):
            return {"status": "error", "error": "test_error"}

    class NoContentResponse:
        def __init__(self):
            self.status_code = 204

        def json(self):
            raise requests.exceptions.JSONDecodeError("No content")


class PrometheusAPIClientTest(PrometheusAPIClientTestBase):
    def test_get(self):
        url = "test"
        root_path = "root_path"
        expected_url = f"http://localhost:9090/{root_path}/api/v1/{url}"

        params = {"query": "ceilometer_image_size{publisher='localhost'}"}
        expected_params = params

        return_value = self.GoodResponse()
        with mock.patch.object(requests.Session, 'get',
                               return_value=return_value) as m:
            c = client.PrometheusAPIClient("localhost:9090",
                                           root_path=root_path)
            c._get(url, params)

        m.assert_called_with(expected_url,
                             params=expected_params,
                             headers={'Accept': 'application/json',
                                      'Accept-Encoding': 'identity'})

    def test_get_error(self):
        url = "test"
        params = {"query": "ceilometer_image_size{publisher='localhost'}"}

        return_value = self.BadResponse()
        with mock.patch.object(requests.Session, 'get',
                               return_value=return_value):
            c = client.PrometheusAPIClient("localhost:9090")
            self.assertRaises(client.PrometheusAPIClientError,
                              c._get, url, params)

        return_value = self.NoContentResponse()
        with mock.patch.object(requests.Session, 'get',
                               return_value=return_value):
            c = client.PrometheusAPIClient("localhost:9090")
            self.assertRaises(client.PrometheusAPIClientError,
                              c._get, url, params)

    def test_post(self):
        url = "test"
        expected_url = f"http://localhost:9090/api/v1/{url}"

        params = {"query": "ceilometer_image_size{publisher='localhost'}"}
        expected_params = params

        return_value = self.GoodResponse()
        with mock.patch.object(requests.Session, 'post',
                               return_value=return_value) as m:
            c = client.PrometheusAPIClient("localhost:9090")
            c._post(url, params)

        m.assert_called_with(expected_url,
                             params=expected_params,
                             headers={'Accept': 'application/json'})

    def test_post_error(self):
        url = "test"
        params = {"query": "ceilometer_image_size{publisher='localhost'}"}

        return_value = self.BadResponse()
        with mock.patch.object(requests.Session, 'post',
                               return_value=return_value):
            c = client.PrometheusAPIClient("localhost:9090")
            self.assertRaises(client.PrometheusAPIClientError,
                              c._post, url, params)

        return_value = self.NoContentResponse()
        with mock.patch.object(requests.Session, 'post',
                               return_value=return_value):
            c = client.PrometheusAPIClient("localhost:9090")
            self.assertRaises(client.PrometheusAPIClientError,
                              c._post, url, params)


class PrometheusAPIClientQueryTest(PrometheusAPIClientTestBase):
    def setUp(self):
        super().setUp()

    class GoodQueryResponse(PrometheusAPIClientTestBase.GoodResponse):
        def __init__(self):
            super().__init__()
            self.result1 = {
                "metric": {
                    "__name__": "test1",
                },
                "value": [103254, "1"]
            }
            self.result2 = {
                "metric": {
                    "__name__": "test2",
                },
                "value": [103255, "2"]
            }
            self.expected = [client.PrometheusMetric(self.result1),
                             client.PrometheusMetric(self.result2)]

        def json(self):
            return {
                "status": "success",
                "data": {
                    "resultType": "vector",
                    "result": [self.result1, self.result2]
                }
            }

    class EmptyQueryResponse(PrometheusAPIClientTestBase.GoodResponse):
        def __init__(self):
            super().__init__()
            self.expected = []

        def json(self):
            return {
                "status": "success",
                "data": {
                    "resultType": "vector",
                    "result": []
                }
            }

    def test_query(self):
        query = "ceilometer_image_size{publisher='localhost.localdomain'}"

        matcher = MetricListMatcher(self.GoodQueryResponse().expected)
        return_value = self.GoodQueryResponse().json()
        with mock.patch.object(client.PrometheusAPIClient, '_get',
                               return_value=return_value) as m:
            c = client.PrometheusAPIClient("localhost:9090")
            ret = c.query(query)

            m.assert_called_with("query", {"query": query})
            self.assertThat(ret, matcher)

        return_value = self.EmptyQueryResponse().json()
        with mock.patch.object(client.PrometheusAPIClient, '_get',
                               return_value=return_value) as m:
            c = client.PrometheusAPIClient("localhost:9090")
            ret = c.query(query)

            self.assertEqual(self.EmptyQueryResponse().expected, ret)

    def test_query_error(self):
        query = "ceilometer_image_size{publisher='localhost.localdomain'}"
        client_exception = client.PrometheusAPIClientError(self.BadResponse())

        with mock.patch.object(client.PrometheusAPIClient, '_get',
                               side_effect=client_exception):
            c = client.PrometheusAPIClient("localhost:9090")

            self.assertRaises(client.PrometheusAPIClientError, c.query, query)


class PrometheusAPIClientSeriesTest(PrometheusAPIClientTestBase):
    def setUp(self):
        super().setUp()

    class GoodSeriesResponse(PrometheusAPIClientTestBase.GoodResponse):
        def __init__(self):
            super().__init__()
            self.data = [{
                "__name__": "up",
                "job": "prometheus",
                "instance": "localhost:9090"
            }, {
                "__name__": "up",
                "job": "node",
                "instance": "localhost:9091"
            }, {
                "__name__": "process_start_time_seconds",
                "job": "prometheus",
                "instance": "localhost:9090"
            }]
            self.expected = self.data

        def json(self):
            return {
                "status": "success",
                "data": self.data
            }

    class EmptySeriesResponse(PrometheusAPIClientTestBase.GoodResponse):
        def __init__(self):
            super().__init__()
            self.data = []
            self.expected = self.data

        def json(self):
            return {
                "status": "success",
                "data": self.data
            }

    def test_series(self):
        matches = ["up", "ceilometer_image_size"]

        return_value = self.GoodSeriesResponse().json()
        with mock.patch.object(client.PrometheusAPIClient, '_get',
                               return_value=return_value) as m:
            c = client.PrometheusAPIClient("localhost:9090")
            ret = c.series(matches)

            m.assert_called_with("series", {"match[]": matches})
            self.assertEqual(ret, self.GoodSeriesResponse().data)

        return_value = self.EmptySeriesResponse().json()
        with mock.patch.object(client.PrometheusAPIClient, '_get',
                               return_value=return_value) as m:
            c = client.PrometheusAPIClient("localhost:9090")
            ret = c.series(matches)

            m.assert_called_with("series", {"match[]": matches})
            self.assertEqual(ret, self.EmptySeriesResponse().data)

    def test_series_error(self):
        matches = ["up", "ceilometer_image_size"]
        client_exception = client.PrometheusAPIClientError(self.BadResponse())

        with mock.patch.object(client.PrometheusAPIClient, '_get',
                               side_effect=client_exception):
            c = client.PrometheusAPIClient("localhost:9090")

            self.assertRaises(client.PrometheusAPIClientError,
                              c.series,
                              matches)


class PrometheusAPIClientLabelsTest(PrometheusAPIClientTestBase):
    def setUp(self):
        super().setUp()

    class GoodLabelsResponse(PrometheusAPIClientTestBase.GoodResponse):
        def __init__(self):
            super().__init__()
            self.labels = ["up", "job", "project_id"]

        def json(self):
            return {
                "status": "success",
                "data": self.labels
            }

    def test_labels(self):
        return_value = self.GoodLabelsResponse().json()
        with mock.patch.object(client.PrometheusAPIClient, '_get',
                               return_value=return_value) as m:
            c = client.PrometheusAPIClient("localhost:9090")
            ret = c.labels()

            m.assert_called_with("labels")
            self.assertEqual(ret, self.GoodLabelsResponse().labels)

    def test_labels_error(self):
        client_exception = client.PrometheusAPIClientError(self.BadResponse())
        with mock.patch.object(client.PrometheusAPIClient, '_get',
                               side_effect=client_exception):
            c = client.PrometheusAPIClient("localhost:9090")

            self.assertRaises(client.PrometheusAPIClientError, c.labels)


class PrometheusAPIClientLabelValuesTest(PrometheusAPIClientTestBase):
    def setUp(self):
        super().setUp()

    class GoodLabelValuesResponse(PrometheusAPIClientTestBase.GoodResponse):
        def __init__(self):
            super().__init__()
            self.values = ["prometheus", "some_other_value"]

        def json(self):
            return {
                "status": "success",
                "data": self.values
            }

    class EmptyLabelValuesResponse(PrometheusAPIClientTestBase.GoodResponse):
        def __init__(self):
            super().__init__()
            self.values = []

        def json(self):
            return {
                "status": "success",
                "data": self.values
            }

    def test_label_values(self):
        label_name = "job"

        return_value = self.GoodLabelValuesResponse().json()
        with mock.patch.object(client.PrometheusAPIClient, '_get',
                               return_value=return_value) as m:
            c = client.PrometheusAPIClient("localhost:9090")
            ret = c.label_values(label_name)

            m.assert_called_with(f"label/{label_name}/values")
            self.assertEqual(ret, self.GoodLabelValuesResponse().values)

        return_value = self.EmptyLabelValuesResponse().json()
        with mock.patch.object(client.PrometheusAPIClient, '_get',
                               return_value=return_value) as m:
            c = client.PrometheusAPIClient("localhost:9090")
            ret = c.label_values(label_name)

            m.assert_called_with(f"label/{label_name}/values")
            self.assertEqual(ret, self.EmptyLabelValuesResponse().values)

    def test_label_values_error(self):
        label_name = "job"
        client_exception = client.PrometheusAPIClientError(self.BadResponse())

        with mock.patch.object(client.PrometheusAPIClient, '_get',
                               side_effect=client_exception):
            c = client.PrometheusAPIClient("localhost:9090")

            self.assertRaises(client.PrometheusAPIClientError,
                              c.label_values,
                              label_name)


class PrometheusAPIClientDeleteTest(PrometheusAPIClientTestBase):
    def setUp(self):
        super().setUp()

    class GoodDeleteResponse(PrometheusAPIClientTestBase.NoContentResponse):
        pass

    def test_delete(self):
        matches = ["{job='prometheus'}", "up"]
        start = 1
        end = 12
        resp = self.GoodDeleteResponse()
        post_exception = client.PrometheusAPIClientError(resp)

        with mock.patch.object(client.PrometheusAPIClient, '_post',
                               side_effect=post_exception) as m:
            c = client.PrometheusAPIClient("localhost:9090")
            # _post is expected to raise an exception. It's expected
            # that the exception is caught inside delete. This
            # test should run without exception getting out of delete
            try:
                c.delete(matches, start, end)
            except Exception as ex:  # noqa: B902
                self.fail("Exception risen by delete: " + ex)

            m.assert_called_with("admin/tsdb/delete_series",
                                 {"match[]": matches,
                                  "start": start,
                                  "end": end})

    def test_delete_error(self):
        matches = ["{job='prometheus'}", "up"]
        client_exception = client.PrometheusAPIClientError(self.BadResponse())

        with mock.patch.object(client.PrometheusAPIClient, '_post',
                               side_effect=client_exception):
            c = client.PrometheusAPIClient("localhost:9090")

            self.assertRaises(client.PrometheusAPIClientError,
                              c.delete,
                              matches)


class PrometheusAPIClientCleanTombstonesTest(PrometheusAPIClientTestBase):
    def setUp(self):
        super().setUp()

    class GoodCleanTombResponse(PrometheusAPIClientTestBase.NoContentResponse):
        pass

    def test_clean_tombstones(self):
        resp = self.GoodCleanTombResponse()
        post_exception = client.PrometheusAPIClientError(resp)

        with mock.patch.object(client.PrometheusAPIClient, '_post',
                               side_effect=post_exception) as m:
            c = client.PrometheusAPIClient("localhost:9090")
            # _post is expected to raise an exception. It's expected
            # that the exception is caught inside clean_tombstones. This
            # test should run without exception getting out of clean_tombstones
            try:
                c.clean_tombstones()
            except Exception as ex:  # noqa: B902
                self.fail("Exception risen by clean_tombstones: " + ex)

            m.assert_called_with("admin/tsdb/clean_tombstones")

    def test_snapshot_error(self):
        client_exception = client.PrometheusAPIClientError(self.BadResponse())

        with mock.patch.object(client.PrometheusAPIClient, '_post',
                               side_effect=client_exception):
            c = client.PrometheusAPIClient("localhost:9090")

            self.assertRaises(client.PrometheusAPIClientError,
                              c.clean_tombstones)


class PrometheusAPIClientSnapshotTest(PrometheusAPIClientTestBase):
    def setUp(self):
        super().setUp()

    class GoodSnapshotResponse(PrometheusAPIClientTestBase.NoContentResponse):
        def __init__(self):
            super().__init__()
            self.filename = "somefilename"

        def json(self):
            return {
                "status": "success",
                "data": {
                    "name": self.filename
                }
            }

    def test_snapshot(self):
        return_value = self.GoodSnapshotResponse().json()
        with mock.patch.object(client.PrometheusAPIClient, '_post',
                               return_value=return_value) as m:
            c = client.PrometheusAPIClient("localhost:9090")
            ret = c.snapshot()

            m.assert_called_with("admin/tsdb/snapshot")
            self.assertEqual(ret, self.GoodSnapshotResponse().filename)

    def test_snapshot_error(self):
        client_exception = client.PrometheusAPIClientError(self.BadResponse())

        with mock.patch.object(client.PrometheusAPIClient, '_post',
                               side_effect=client_exception):
            c = client.PrometheusAPIClient("localhost:9090")

            self.assertRaises(client.PrometheusAPIClientError,
                              c.snapshot)



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import keystoneauth1.session

from observabilityclient import rbac
from observabilityclient.utils.metric_utils import get_prometheus_client
from observabilityclient.v1 import python_api


class Client:
    """Client for the observabilityclient api."""

    def __init__(self, session=None, adapter_options=None,
                 session_options=None, disable_rbac=False):
        """Initialize a new client for the Observabilityclient v1 API."""
        session_options = session_options or {}
        adapter_options = adapter_options or {}

        adapter_options.setdefault('service_type', "metric-storage")

        if session is None:
            session = keystoneauth1.session.Session(**session_options)
        else:
            if session_options:
                raise ValueError("session and session_options are exclusive")

        self.session = session

        self.prometheus_client = get_prometheus_client(
            session, adapter_options
        )
        self.query = python_api.QueryManager(self)
        self.rbac = rbac.PromQLRbac(
            self.prometheus_client,
            self.session.get_project_id()
        )



#   Copyright 2022 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import argparse
import sys

from osc_lib.command import command

from observabilityclient.i18n import _


class DeprecatedNOOPAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        kwargs['help'] = f"[DEPRECATED] {kwargs['help']}" \
                         if 'help' in kwargs else \
                         "[DEPRECATED]"
        kwargs['required'] = False
        kwargs['nargs'] = 0
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string is not None:
            print(f"Note: {option_string} is deprecated option, "
                  "has no effect and is a subject for removal "
                  "in future OpenStack release.", file=sys.stderr)


class ObservabilityBaseCommand(command.Command):
    """Base class for metric commands."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--disable-rbac',
            action=DeprecatedNOOPAction,
            default=True,
            help=_("Disable rbac injection")
        )
        return parser


class Manager:
    """Base class for the python api."""

    DEFAULT_HEADERS = {
        "Accept": "application/json",
    }

    def __init__(self, client):
        self.client = client
        self.prom = client.prometheus_client

    def _set_default_headers(self, kwargs):
        headers = kwargs.get('headers', {})
        for k, v in self.DEFAULT_HEADERS.items():
            if k not in headers:
                headers[k] = v
        kwargs['headers'] = headers
        return kwargs

    def _get(self, *args, **kwargs):
        self._set_default_headers(kwargs)
        return self.client.api.get(*args, **kwargs)

    def _post(self, *args, **kwargs):
        self._set_default_headers(kwargs)
        return self.client.api.post(*args, **kwargs)

    def _put(self, *args, **kwargs):
        self._set_default_headers(kwargs)
        return self.client.api.put(*args, **kwargs)

    def _patch(self, *args, **kwargs):
        self._set_default_headers(kwargs)
        return self.client.api.patch(*args, **kwargs)

    def _delete(self, *args, **kwargs):
        self._set_default_headers(kwargs)
        return self.client.api.delete(*args, **kwargs)



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import re

from observabilityclient.utils.metric_utils import format_labels


class PromQLRbac:
    def __init__(self, prom_api_client, project_id, project_label='project'):
        self.client = prom_api_client

        # NOTE(jwysogla): Since Prometheus 3, metric and label names can
        # utilize all unicode characters. But the syntax is a little different
        # and some parts of the queries need to be quoted or escaped. The
        # rest of this module doesn't support that right now, so use a
        # Prometheus 2 regex and raise an exception when it doesn't match.
        # See https://prometheus.io/docs/concepts/data_model/
        label_name_regex = "[a-zA-Z_][a-zA-Z0-9_]*"
        if not re.fullmatch(label_name_regex, project_label):
            raise ValueError(
                f"Project label {project_label} doesn't match the "
                f"label name regex: {label_name_regex}"
            )

        self.labels = {
            project_label: project_id
        }

    def _find_label_value_end(self, query, start, quote_char):
        end = start
        while (end == start or query[end - 1] == '\\'):
            # Looking for first unescaped quotes
            end = query.find(quote_char, end + 1)
        # returns the quote position or -1
        return end

    def _find_match_operator(self, query, start):
        eq_sign_pos = query.find('=', start)
        tilde_pos = query.find('~', start)
        if eq_sign_pos == -1:
            return tilde_pos
        if tilde_pos == -1:
            return eq_sign_pos
        return min(eq_sign_pos, tilde_pos)

    def _find_label_pair_end(self, query, start):
        match_operator_pos = self._find_match_operator(query, start)
        quote_char = "'"
        quote_start_pos = query.find(quote_char, match_operator_pos)
        if quote_start_pos == -1:
            quote_char = '"'
            quote_start_pos = query.find(quote_char, match_operator_pos)
        end = self._find_label_value_end(query, quote_start_pos, quote_char)
        # returns the pair end or -1
        return end

    def _find_label_section_end(self, query, start):
        if query[start:].startswith('{}'):
            # We have an empty label section, without any
            # label pair
            return start + 1
        nearest_curly_brace_pos = None
        while nearest_curly_brace_pos != -1:
            pair_end = self._find_label_pair_end(query, start)
            nearest_curly_brace_pos = query.find("}", pair_end)
            nearest_match_operator_pos = self._find_match_operator(query,
                                                                   pair_end)
            if (nearest_curly_brace_pos < nearest_match_operator_pos or
                    nearest_match_operator_pos == -1):
                # If we have "}" before the nearest "=" or "~",
                # then we must be at the end of the label section
                # and the "=" or "~" is a part of the next section.
                return nearest_curly_brace_pos
            start = pair_end
        # TODO(tkajinam): We should probably raise an exception here because
        # this indicates illegal format without closing } .
        return -1

    def _insert_labels(self, query, location, comma=False, braces=False):
        comma_str = ", " if comma else ""
        formatted_labels = format_labels(self.labels)
        labels_str = f"{{{formatted_labels}}}" if braces else formatted_labels
        return (f"{query[:location]}{comma_str}"
                f"{labels_str}"
                f"{query[location:]}")

    def modify_query(self, query, metric_names=None):
        """Add rbac labels to a query.

        :param query: The query to modify
        :type query: str

        :param metric_names: List of metric names currently stored in
                             prometheus. For correct function of the query
                             modification, it's important for the list to be
                             accurate and to include all metrics across all
                             tenants. This parameter can be unspecified or
                             None, in which case the list will be retrieved
                             from Prometheus by sending an API request to it.
                             It's advised to provide the metric list when
                             doing a query modification in a loop for
                             performance purposes.
        :type metric_names: list
        """

        if metric_names is None:
            metric_names = self.client.label_values("__name__")

        # We need to detect the locations of metric names
        # inside the query
        # NOTE the locations are the locations within the original query
        name_end_locations = []
        for name in metric_names:
            # Regex for a metric name is: [a-zA-Z_:][a-zA-Z0-9_:]*
            # We need to make sure, that "name" isn't just a part
            # of a longer word, so we try to expand it by "name_regex"
            name_regex = "[a-zA-Z_:]?[a-zA-Z0-9_:]*" + name + "[a-zA-Z0-9_:]*"
            potential_names = re.finditer(name_regex, query)
            for potential_name in potential_names:
                if potential_name.group(0) == name:
                    name_end_locations.append(potential_name.end())

        name_end_locations = sorted(name_end_locations, reverse=True)
        if len(name_end_locations) == 0:
            name_end_locations = [0]
        for name_end_location in name_end_locations:
            if (name_end_location < len(query) and
                    query[name_end_location] == "{"):
                # There is already a label section
                labels_end = self._find_label_section_end(
                    query,
                    name_end_location
                )
                if labels_end == name_end_location + 1:
                    query = self._insert_labels(
                        query,
                        labels_end,
                        comma=False,
                        braces=False
                    )
                else:
                    query = self._insert_labels(
                        query,
                        labels_end,
                        comma=True,
                        braces=False
                    )
            else:
                query = self._insert_labels(
                    query,
                    name_end_location,
                    comma=False,
                    braces=True
                )
        return query

    def append_rbac_labels(self, query):
        """Append rbac labels to queries.

        It's a simplified and faster version of modify_query(). This just
        appends the labels at the end of the query string. For proper handling
        of complex queries, where metric names might occure elsewhere than
        just at the end, please use the modify_query() function.

        :param query: The query to append to
        :type query: str
        """
        if any(c in query for c in "{}"):
            return self.modify_query(query)
        else:
            return f"{query}{{{format_labels(self.labels)}}}"



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

import testtools

from observabilityclient import rbac


class PromQLRbacTest(testtools.TestCase):
    def setUp(self):
        super().setUp()
        self.project_id = "project123"
        self.rbac = rbac.PromQLRbac(mock.Mock(), mock.Mock())
        self.rbac.labels = {
            "project": "project123"
        }
        self.rbac.client.label_values = mock.MagicMock(
            return_value=[
                'test_query',
                'cpu_temp_celsius',
                'http_requests',
                'test:query:with:colon:',
                'test_query_with_digit1',
                'method_code:http_errors:rate5m',
                'method:http_requests:rate5m'
            ]
        )
        self.test_cases = [
            (
                "test_query",
                f"test_query{{project='{self.project_id}'}}"
            ), (
                "test_query{somelabel='value'}",

                (f"test_query{{somelabel='value', "
                 f"project='{self.project_id}'}}")
            ), (
                "test_query{somelabel='value', label2='value2'}",

                (f"test_query{{somelabel='value', label2='value2', "
                 f"project='{self.project_id}'}}")
            ), (
                "test_query{somelabel='unicode{}{ \t/-_#~$&%\\'}",

                (f"test_query{{somelabel='unicode{{}}{{ \t/-_#~$&%\\', "
                 f"project='{self.project_id}'}}")
            ), (
                "test_query{somelabel='s p a c e'}",

                (f"test_query{{somelabel='s p a c e', "
                 f"project='{self.project_id}'}}")
            ), (
                "test_query{somelabel='doublequotes\"'}",

                (f"test_query{{somelabel='doublequotes\"', "
                 f"project='{self.project_id}'}}")
            ), (
                'test_query{somelabel="singlequotes\'"}',

                (f'test_query{{somelabel="singlequotes\'", '
                 f'project=\'{self.project_id}\'}}')
            ), (
                "test_query{doesnt_match_regex!~'regex'}",

                (f"test_query{{doesnt_match_regex!~'regex', "
                 f"project='{self.project_id}'}}")
            ), (
                "delta(cpu_temp_celsius{host='zeus'}[2h]) - "
                "sum(http_requests) + "
                "sum(http_requests{instance=~'.*'}) + "
                "sum(http_requests{or_regex=~'smth1|something2|3'})",

                (f"delta(cpu_temp_celsius{{host='zeus', "
                 f"project='{self.project_id}'}}[2h]) - "
                 f"sum(http_requests"
                 f"{{project='{self.project_id}'}}) + "
                 f"sum(http_requests{{instance=~'.*', "
                 f"project='{self.project_id}'}}) + "
                 f"sum(http_requests{{or_regex=~'smth1|something2|3', "
                 f"project='{self.project_id}'}})")
            ), (
                "round(test_query{label='something'},5)",

                (f"round(test_query{{label='something', "
                 f"project='{self.project_id}'}},5)")
            ), (
                "sum by (foo) (test_query{label_1='baz'})",

                (f"sum by (foo) (test_query{{label_1='baz', "
                 f"project='{self.project_id}'}})")
            ), (
                "test_query{} + avg without (application, group) "
                "(test:query:with:colon:{label='baz'})",

                (f"test_query{{project='{self.project_id}'}} + "
                 f"avg without (application, group) "
                 f"(test:query:with:colon:{{label='baz', "
                 f"project='{self.project_id}'}})")
            ), (
                "test_query{label1='foo'} + on (label1,label2) "
                "avg by (label3) (test_query_with_digit1{label='baz',"
                "label1='foo',label2='bar'})",

                (f"test_query{{label1='foo', "
                 f"project='{self.project_id}'}} "
                 f"+ on (label1,label2) avg by (label3) "
                 f"(test_query_with_digit1{{label='baz',"
                 f"label1='foo',label2='bar', "
                 f"project='{self.project_id}'}})")
            ), (
                "{label='no-metric'}",

                (f"{{label='no-metric', "
                 f"project='{self.project_id}'}}")
            ), (
                "http_requests{environment=~"
                "'staging|testing|development',method!='GET'}",

                (f"http_requests{{environment=~"
                 f"'staging|testing|development',method!='GET', "
                 f"project='{self.project_id}'}}")
            ), (
                "http_requests{replica!='rep-a',replica=~'rep.*'}",

                (f"http_requests{{replica!='rep-a',replica=~'rep.*', "
                 f"project='{self.project_id}'}}")
            ), (
                "{__name__=~'job:.*'}",

                (f"{{__name__=~'job:.*', "
                 f"project='{self.project_id}'}}")
            ), (
                "http_requests offset 5m",

                (f"http_requests"
                 f"{{project='{self.project_id}'}} "
                 f"offset 5m")
            ), (
                "rate(http_requests[5m] offset -1w)",

                (f"rate(http_requests"
                 f"{{project='{self.project_id}'}}"
                 f"[5m] offset -1w)")
            ), (
                "http_requests @ 1609746000",

                (f"http_requests"
                 f"{{project='{self.project_id}'}} "
                 f"@ 1609746000")
            ), (
                "histogram_quantile(0.9, sum by (le) "
                "(rate(http_requests[10m])))",

                (f"histogram_quantile(0.9, sum by (le) "
                 f"(rate(http_requests"
                 f"{{project='{self.project_id}'}}"
                 f"[10m])))"
                 )
            ), (
                "test_query{project='some_id'}",

                (f"test_query{{project='some_id', "
                 f"project='{self.project_id}'}}"
                 )
            )
        ]

    def test_constructor(self):
        r = rbac.PromQLRbac("client", "123")
        self.assertEqual(r.labels, {
            "project": "123"
        })

    def test_modify_query(self):
        for query, expected in self.test_cases:
            ret = self.rbac.modify_query(query)
            self.assertEqual(expected, ret)

    def test_append_rbac_labels(self):
        query = "test_query"
        expected = f"{query}{{project='{self.project_id}'}}"
        ret = self.rbac.append_rbac_labels(query)
        self.assertEqual(expected, ret)

    def test_setting_different_project_label_name(self):
        query = 'test_query'
        project_label = 'different_name'
        project_value = 'some_project'

        rbac_instance = rbac.PromQLRbac(
            mock.Mock(), project_value, project_label=project_label
        )
        rbac_instance.client.label_values = lambda label: [
            query,
        ]

        expected = f"{query}{{{project_label}='{project_value}'}}"

        ret = rbac_instance.modify_query(query)

        self.assertEqual(expected, ret)

    def test_setting_illegal_project_label_name(self):
        # Try setting a label name with a white space character inside
        project_label = 'different name'
        project_value = 'some_project'

        args = (mock.Mock(), project_value)
        kwargs = {'project_label': project_label}

        self.assertRaises(ValueError, rbac.PromQLRbac, *args, **kwargs)

        # Try setting an empty label name
        project_label = ''

        kwargs = {'project_label': project_label}

        self.assertRaises(ValueError, rbac.PromQLRbac, *args, **kwargs)

    def test_specifying_metric_names(self):
        query = "test_query"
        metric_names = [
            "metric_name1",
            "foo",
            "bar",
            "test_query"
        ]

        expected = "test_query{project='project123'}"

        ret = self.rbac.modify_query(query, metric_names)

        self.assertEqual(expected, ret)
        self.rbac.client.label_values.assert_not_called()

    def test_not_specifying_metric_names(self):
        query = "test_query"

        expected = "test_query{project='project123'}"

        ret = self.rbac.modify_query(query)

        self.assertEqual(expected, ret)
        self.rbac.client.label_values.assert_called_once()



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from cliff import lister

from observabilityclient.i18n import _
from observabilityclient.utils import metric_utils
from observabilityclient.v1 import base


class List(base.ObservabilityBaseCommand, lister.Lister):
    """Query prometheus for list of all metrics."""

    def take_action(self, parsed_args):
        client = metric_utils.get_client(self)
        metrics = client.query.list(disable_rbac=parsed_args.disable_rbac)
        return ["metric_name"], [[m] for m in metrics]


class Show(base.ObservabilityBaseCommand, lister.Lister):
    """Query prometheus for the current value of metric."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            help=_("Name of the metric to show"))
        return parser

    def take_action(self, parsed_args):
        client = metric_utils.get_client(self)
        metric = client.query.show(parsed_args.name,
                                   disable_rbac=parsed_args.disable_rbac)
        ret = metric_utils.metrics2cols(metric)
        return ret


class Query(base.ObservabilityBaseCommand, lister.Lister):
    """Query prometheus with a custom query string."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'query',
            help=_("Custom PromQL query"))
        return parser

    def take_action(self, parsed_args):
        client = metric_utils.get_client(self)
        metric = client.query.query(parsed_args.query,
                                    disable_rbac=parsed_args.disable_rbac)
        ret = metric_utils.metrics2cols(metric)
        return ret


class Delete(base.ObservabilityBaseCommand):
    """Delete data for a selected series and time range."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'matches',
            action="append",
            nargs='+',
            help=_("Series selector, that selects the series to delete. "
                   "Specify multiple selectors delimited by space to "
                   "delete multiple series."))
        parser.add_argument(
            '--start',
            help=_("Start timestamp in rfc3339 or unix timestamp. "
                   "Defaults to minimum possible timestamp."))
        parser.add_argument(
            '--end',
            help=_("End timestamp in rfc3339 or unix timestamp. "
                   "Defaults to maximum possible timestamp."))
        return parser

    def take_action(self, parsed_args):
        client = metric_utils.get_client(self)
        return client.query.delete(parsed_args.matches,
                                   parsed_args.start,
                                   parsed_args.end)


class CleanTombstones(base.ObservabilityBaseCommand):
    """Remove deleted data from disk and clean up the existing tombstones."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        client = metric_utils.get_client(self)
        return client.query.clean_tombstones()


class Snapshot(base.ObservabilityBaseCommand, lister.Lister):
    def take_action(self, parsed_args):
        client = metric_utils.get_client(self)
        ret = client.query.snapshot()
        return ["Snapshot file name"], [[ret]]



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from observabilityclient.tests.functional import base

import time


class PythonAPITestFunctionalRBACDisabled(base.PythonAPITestCase):
    def test_list(self):
        ret = self.client.query.list(disable_rbac=True)

        self.assertIn("up", ret)

    def test_show(self):
        ret = self.client.query.show("up", disable_rbac=True)

        for metric in ret:
            self.assertEqual("up", metric.labels["__name__"])
            self.assertEqual("1", metric.value)

    def test_query(self):
        ret = self.client.query.query("up", disable_rbac=True)

        for metric in ret:
            self.assertEqual("up", metric.labels["__name__"])
            self.assertEqual("1", metric.value)


class PythonAPITestFunctionalRBACEnabled(base.PythonAPITestCase):
    def test_list(self):
        ret = self.client.query.list(disable_rbac=False)

        self.assertIn("ceilometer_image_size", ret)
        self.assertNotIn("up", ret)

    def test_show(self):
        ret = self.client.query.show("ceilometer_image_size",
                                     disable_rbac=False)

        for metric in ret:
            self.assertEqual("ceilometer_image_size",
                             metric.labels["__name__"])
            self.assertEqual("custom",
                             metric.labels["job"])

    def test_query(self):
        ret = self.client.query.query("ceilometer_image_size",
                                      disable_rbac=False)

        for metric in ret:
            self.assertEqual("ceilometer_image_size",
                             metric.labels["__name__"])
            self.assertEqual("custom", metric.labels["job"])


class PythonAPITestFunctionalAdminCommands(base.PythonAPITestCase):
    def test_delete(self):
        now = time.time()
        metric_name = "prometheus_build_info"
        query = f"{metric_name}@{now}"

        query_before = self.client.query.query(query, disable_rbac=True)

        for metric in query_before:
            self.assertEqual(metric_name, metric.labels["__name__"])

        self.client.query.delete(metric_name)

        query_after = self.client.query.query(query, disable_rbac=True)
        self.assertEqual([], query_after)

    def test_clean_tombstones(self):
        # NOTE(jwysogla) There is not much to check here
        # except for the fact, that the command doesn't
        # raise an exception. Prometheus doesn't send any
        # data back and we don't have a reliable way to query
        # prometheus that this command did something.
        self.client.query.clean_tombstones()

    def test_snapshot(self):
        ret = self.client.query.snapshot()
        self.assertIn(time.strftime("%Y%m%d"), ret)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# base.py file taken and modified from the openstackclient functional tests

import json
import logging
import os
import shlex
import subprocess

from observabilityclient import client

from keystoneauth1 import loading
from keystoneauth1 import session
from openstack import config as occ
from tempest.lib.cli import output_parser
from tempest.lib import exceptions
import testtools

ADMIN_CLOUD = os.environ.get('OS_ADMIN_CLOUD', 'devstack-admin')
LOG = logging.getLogger(__name__)


class PythonAPITestCase(testtools.TestCase):
    def _getKeystoneSession(self):
        conf = occ.OpenStackConfig()
        creds = conf.get_one(cloud=ADMIN_CLOUD).get_auth_args()
        ks_creds = dict(
            auth_url=creds["auth_url"],
            username=creds["username"],
            password=creds["password"],
            project_name=creds["project_name"],
            user_domain_id=creds["user_domain_id"],
            project_domain_id=creds["project_domain_id"])
        loader = loading.get_plugin_loader("password")
        auth = loader.load_from_options(**ks_creds)
        return session.Session(auth=auth)

    def setUp(self):
        super().setUp()
        self.client = client.Client(
            1,
            self._getKeystoneSession()
        )


def execute(cmd, fail_ok=False, merge_stderr=False):
    """Execute specified command for the given action."""
    LOG.debug('Executing: %s', cmd)
    cmdlist = shlex.split(cmd)
    stdout = subprocess.PIPE
    stderr = subprocess.STDOUT if merge_stderr else subprocess.PIPE

    proc = subprocess.Popen(cmdlist, stdout=stdout, stderr=stderr)

    result_out, result_err = proc.communicate()
    result_out = result_out.decode('utf-8')
    LOG.debug('stdout: %s', result_out)
    LOG.debug('stderr: %s', result_err)

    if not fail_ok and proc.returncode != 0:
        raise exceptions.CommandFailed(
            proc.returncode,
            cmd,
            result_out,
            result_err,
        )

    return result_out


class CliTestCase(testtools.TestCase):
    @classmethod
    def openstack(
        cls,
        cmd,
        *,
        cloud=ADMIN_CLOUD,
        fail_ok=False,
        parse_output=False,
    ):
        """Execute observabilityclient command for the given action.

        :param cmd: A string representation of the command to execute.
        :param cloud: The cloud to execute against. This can be a string, empty
            string, or None. A string results in '--os-auth-type $cloud', an
            empty string results in the '--os-auth-type' option being
            omitted, and None resuts in '--os-auth-type none' for legacy
            reasons.
        :param fail_ok: If failure is permitted. If False (default), a command
            failure will result in `~tempest.lib.exceptions.CommandFailed`
            being raised.
        :param parse_output: If true, pass the '-f json' parameter and decode
            the output.
        :returns: The output from the command.
        :raises: `~tempest.lib.exceptions.CommandFailed` if the command failed
            and ``fail_ok`` was ``False``.
        """
        auth_args = []
        if cloud is None:
            # Execute command with no auth
            auth_args.append('--os-auth-type none')
        elif cloud != '':
            # Execute command with an explicit cloud specified
            auth_args.append(f'--os-cloud {cloud}')

        format_args = []
        if parse_output:
            format_args.append('-f json')

        output = execute(
            ' '.join(['openstack'] + auth_args + [cmd] + format_args),
            fail_ok=fail_ok,
        )

        if parse_output:
            return json.loads(output)
        else:
            return output

    @classmethod
    def assertOutput(cls, expected, actual):
        if expected != actual:
            raise Exception(expected + ' != ' + actual)

    @classmethod
    def assertInOutput(cls, expected, actual):
        if expected not in actual:
            raise Exception(expected + ' not in ' + actual)

    @classmethod
    def assertNotInOutput(cls, expected, actual):
        if expected in actual:
            raise Exception(expected + ' in ' + actual)

    @classmethod
    def assertsOutputNotNone(cls, observed):
        if observed is None:
            raise Exception('No output observed')

    def assert_table_structure(self, items, field_names):
        """Verify that all items have keys listed in field_names."""
        for item in items:
            for field in field_names:
                self.assertIn(field, item)

    def assert_show_fields(self, show_output, field_names):
        """Verify that all items have keys listed in field_names."""
        # field_names = ['name', 'description']
        # show_output = [{'name': 'fc2b98d8faed4126b9e371eda045ade2'},
        #          {'description': 'description-821397086'}]
        # this next line creates a flattened list of all 'keys' (like 'name',
        # and 'description' out of the output
        all_headers = [item for sublist in show_output for item in sublist]
        for field_name in field_names:
            self.assertIn(field_name, all_headers)

    def parse_show_as_object(self, raw_output):
        """Return a dict with values parsed from cli output."""
        items = self.parse_show(raw_output)
        o = {}
        for item in items:
            o.update(item)
        return o

    def parse_show(self, raw_output):
        """Return list of dicts with item values parsed from cli output."""
        items = []
        table_ = output_parser.table(raw_output)
        for row in table_['values']:
            item = {}
            item[row[0]] = row[1]
            items.append(item)
        return items

    def parse_listing(self, raw_output):
        """Return list of dicts with basic item parsed from cli output."""
        return output_parser.listing(raw_output)



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from observabilityclient.utils.metric_utils import format_labels
from observabilityclient.v1 import base


class QueryManager(base.Manager):
    def list(self, disable_rbac=True):
        """List metric names.

        :param disable_rbac: Disables rbac injection if set to True
        :type disable_rbac: boolean
        """
        if disable_rbac:
            metric_names = self.prom.label_values("__name__")
            return sorted(metric_names)
        else:
            match = f"{{{format_labels(self.client.rbac.labels)}}}"
            metrics = self.prom.series(match)
            if metrics == []:
                return []
            unique_metric_names = list({m['__name__'] for m in metrics})
            return sorted(unique_metric_names)

    def show(self, name, disable_rbac=True):
        """Show current values for metrics of a specified name.

        :param disable_rbac: Disables rbac injection if set to True
        :type disable_rbac: boolean
        """
        query = ""
        if disable_rbac:
            query = name
        else:
            query = self.client.rbac.append_rbac_labels(name)
        last_metric_query = f"last_over_time({query}[5m])"
        return self.prom.query(last_metric_query)

    def query(self, query, disable_rbac=True):
        """Send a query to prometheus.

        The query can be any PromQL query. Labels for enforcing
        rbac will be added to all of the metric name inside the query.
        Having labels as part of a query is allowed.

        A call like this:
        query("sum(name1) - sum(name2{label1='value'})")
        will result in a query string like this:
        "sum(name1{rbac='rbac_value'}) -
        sum(name2{label1='value', rbac='rbac_value'})"

        :param query: Custom query string
        :type query: str
        :param disable_rbac: Disables rbac injection if set to True
        :type disable_rbac: boolean
        """
        if not disable_rbac:
            query = self.client.rbac.modify_query(query)
        return self.prom.query(query)

    def delete(self, matches, start=None, end=None):
        """Delete metrics from Prometheus.

        The metrics aren't deleted immediately. Do a call to clean_tombstones()
        to speed up the deletion. If start and end isn't specified, then
        minimum and maximum timestamps are used.

        :param matches: List of matches to match which metrics to delete
        :type matches: [str]
        :param start: timestamp from which to start deleting
        :type start: rfc3339 or unix_timestamp
        :param end: timestamp until which to delete
        :type end: rfc3339 or unix_timestamp
        """
        # TODO(jwysogla) Do we want to restrict access to the admin api
        #                endpoints? We could either try to inject
        #                the project label like in query. We could also
        #                do some check right here, before
        #                it gets to prometheus.
        return self.prom.delete(matches, start, end)

    def clean_tombstones(self):
        """Instruct prometheus to clean tombstones."""
        return self.prom.clean_tombstones()

    def snapshot(self):
        """Create a snapshot of the current data."""
        return self.prom.snapshot()



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

import testtools

from observabilityclient import prometheus_client
from observabilityclient import rbac
from observabilityclient.tests.unit.test_prometheus_client import (
    MetricListMatcher
)
from observabilityclient.v1 import python_api


class QueryManagerTest(testtools.TestCase):
    def setUp(self):
        super().setUp()
        self.client = mock.Mock()
        prom_client = prometheus_client.PrometheusAPIClient("somehost")
        self.client.prometheus_client = prom_client

        self.rbac = mock.Mock(wraps=rbac.PromQLRbac(prom_client, mock.Mock()))
        self.rbac.labels = {'project': 'project_id'}

        self.manager = python_api.QueryManager(self.client)

        self.client.rbac = self.rbac
        self.client.query = self.manager

    def test_list(self):
        returned_data = {'data': ['metric1', 'test42', 'abc2']}
        expected = ['abc2', 'metric1', 'test42']

        with mock.patch.object(prometheus_client.PrometheusAPIClient, '_get',
                               return_value=returned_data):
            ret1 = self.manager.list()
            self.assertEqual(expected, ret1)

            ret2 = self.manager.list(disable_rbac=True)
            self.assertEqual(expected, ret2)

    def test_show(self):
        query = 'some_metric'
        returned_by_prom = {
            'data': {
                'resultType': 'non-vector'
            },
            'value': [1234567, 42],
            'metric': {
                'label': 'label_value'
            }
        }
        expected = [prometheus_client.PrometheusMetric(returned_by_prom)]
        expected_matcher = MetricListMatcher(expected)

        with mock.patch.object(prometheus_client.PrometheusAPIClient, '_get',
                               return_value=returned_by_prom):
            ret2 = self.manager.show(query, disable_rbac=True)
            self.rbac.append_rbac_labels.assert_not_called()

        with mock.patch.object(prometheus_client.PrometheusAPIClient, '_get',
                               return_value=returned_by_prom):
            ret1 = self.manager.show(query)
            self.rbac.append_rbac_labels.assert_not_called()

        self.assertThat(ret1, expected_matcher)
        self.assertThat(ret2, expected_matcher)

    def test_query(self):
        queried_metric_name = 'some_metric'
        query = queried_metric_name
        returned_by_prom = {
            'data': {
                'resultType': 'non-vector'
            },
            'value': [1234567, 42],
            'metric': {
                'label': 'label_value'
            }
        }
        expected = [prometheus_client.PrometheusMetric(returned_by_prom)]
        expected_matcher = MetricListMatcher(expected)
        with mock.patch.object(prometheus_client.PrometheusAPIClient, '_get',
                               return_value=returned_by_prom), \
                mock.patch.object(python_api.QueryManager, 'list',
                                  return_value=queried_metric_name):

            ret1 = self.manager.query(query, disable_rbac=True)
            self.rbac.modify_query.assert_not_called()

            ret2 = self.manager.query(query)
            self.rbac.modify_query.assert_not_called()

        self.assertThat(ret1, expected_matcher)
        self.assertThat(ret2, expected_matcher)

    def test_delete(self):
        matches = "some_metric"
        start = 0
        end = 100
        with mock.patch.object(prometheus_client.PrometheusAPIClient,
                               'delete') as m:
            self.manager.delete(matches, start, end)
        m.assert_called_with(matches, start, end)

    def test_clean_tombstones(self):
        with mock.patch.object(prometheus_client.PrometheusAPIClient,
                               'clean_tombstones') as m:
            self.manager.clean_tombstones()
        m.assert_called_once()

    def test_snapshot(self):
        with mock.patch.object(prometheus_client.PrometheusAPIClient,
                               'snapshot') as m:
            self.manager.snapshot()
        m.assert_called_once()



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

import requests


LOG = logging.getLogger(__name__)


class PrometheusAPIClientError(Exception):
    def __init__(self, response):
        self.resp = response

    def __str__(self) -> str:
        if self.resp.status_code != requests.codes.ok:
            if self.resp.status_code != 204:
                try:
                    decoded = self.resp.json()
                    if 'error' in decoded:
                        return f'[{self.resp.status_code}] {decoded["error"]}'
                except requests.JSONDecodeError:
                    # If an https endpoint is accessed as http,
                    # we get 400 status with plain text instead of
                    # json and decoding it raises exception.
                    return f'[{self.resp.status_code}] {self.resp.text}'
            return f'[{self.resp.status_code}] {self.resp.reason}'
        else:
            decoded = self.resp.json()
            return f'[{decoded.status}]'

    def __repr__(self) -> str:
        return self.__str__()


class PrometheusMetric:
    def __init__(self, input):
        self.timestamp = input['value'][0]
        self.labels = input['metric']
        self.value = input['value'][1]


class PrometheusAPIClient:
    def __init__(self, host, session=None, root_path=""):
        self._host = host
        if not self._host.endswith('/'):
            self._host += '/'
        if session is None:
            self._session = requests.Session()
        else:
            self._session = session
        self._session.verify = False
        self._root_path = root_path
        if root_path != "" and not self._root_path.endswith('/'):
            self._root_path += '/'

    def set_ca_cert(self, ca_cert):
        self._session.verify = ca_cert

    def set_client_cert(self, client_cert, client_key):
        self._session.cert = (client_cert, client_key)

    def set_basic_auth(self, auth_user, auth_password):
        self._session.auth = (auth_user, auth_password)

    def _get_url(self, endpoint):
        scheme = 'https' if self._session.verify else 'http'
        return f"{scheme}://{self._host}{self._root_path}api/v1/{endpoint}"

    def _get(self, endpoint, params=None):
        url = self._get_url(endpoint)
        resp = self._session.get(url, params=params,
                                 headers={'Accept': 'application/json',
                                          'Accept-Encoding': 'identity'})
        if resp.status_code != requests.codes.ok:
            raise PrometheusAPIClientError(resp)
        decoded = resp.json()
        if decoded['status'] != 'success':
            raise PrometheusAPIClientError(resp)

        return decoded

    def _post(self, endpoint, params=None):
        url = self._get_url(endpoint)
        resp = self._session.post(url, params=params,
                                  headers={'Accept': 'application/json'})
        if resp.status_code != requests.codes.ok:
            raise PrometheusAPIClientError(resp)
        decoded = resp.json()
        if 'status' in decoded and decoded['status'] != 'success':
            raise PrometheusAPIClientError(resp)
        return decoded

    def query(self, query):
        """Send custom queries to Prometheus.

        :param query: the query to send
        :type query: str
        """
        LOG.debug("Querying prometheus with query: %s", query)
        decoded = self._get("query", dict(query=query))

        if decoded['data']['resultType'] == 'vector':
            result = [PrometheusMetric(i) for i in decoded['data']['result']]
        else:
            result = [PrometheusMetric(decoded)]
        return result

    def series(self, matches):
        """Query the /series/ endpoint of prometheus.

        :param matches: List of matches to send as parameters
        :type matches: [str]
        """
        LOG.debug("Querying prometheus for series with matches: %s", matches)
        decoded = self._get("series", {"match[]": matches})

        return decoded['data']

    def labels(self):
        """Query the /labels/ endpoint of prometheus, returns list of labels.

        There isn't a way to tell prometheus to restrict
        which labels to return. It's not possible to enforce
        rbac with this for example.
        """
        LOG.debug("Querying prometheus for labels")
        decoded = self._get("labels")

        return decoded['data']

    def label_values(self, label):
        """Query prometheus for values of a specified label.

        :param label: Name of label for which to return values
        :type label: str
        """
        LOG.debug("Querying prometheus for the values of label: %s", label)
        decoded = self._get(f"label/{label}/values")

        return decoded['data']

    # ---------
    # admin api
    # ---------

    def delete(self, matches, start=None, end=None):
        """Delete some metrics from prometheus.

        :param matches: List of matches, that specify which metrics to delete
        :type matches [str]
        :param start: Timestamp from which to start deleting.
                      None for as early as possible.
        :type start: timestamp
        :param end: Timestamp until which to delete.
                    None for as late as possible.
        :type end: timestamp
        """
        # NOTE Prometheus doesn't seem to return anything except
        #      of 204 status code. There doesn't seem to be a
        #      way to know if anything got actually deleted.
        #      It does however return 500 code and error msg
        #      if the admin APIs are disabled.
        LOG.debug("Deleting metrics from prometheus matching: %s", matches)
        try:
            self._post("admin/tsdb/delete_series", {"match[]": matches,
                                                    "start": start,
                                                    "end": end})
        except PrometheusAPIClientError as exc:
            # The 204 is allowed here. 204 is "No Content",
            # which is expected on a successful call
            if exc.resp.status_code != 204:
                raise exc

    def clean_tombstones(self):
        """Ask prometheus to clean tombstones."""
        LOG.debug("Cleaning tombstones from prometheus")
        try:
            self._post("admin/tsdb/clean_tombstones")
        except PrometheusAPIClientError as exc:
            # The 204 is allowed here. 204 is "No Content",
            # which is expected on a successful call
            if exc.resp.status_code != 204:
                raise exc

    def snapshot(self):
        """Create a snapshot and return the file name containing the data."""
        LOG.debug("Taking prometheus data snapshot")
        ret = self._post("admin/tsdb/snapshot")
        return ret["data"]["name"]



#   Copyright 2023 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging
import os
from urllib import parse

from keystoneauth1 import adapter
from keystoneauth1.exceptions import catalog as keystone_exception
from oslo_utils import netutils
import yaml

from observabilityclient.prometheus_client import PrometheusAPIClient


DEFAULT_CONFIG_LOCATIONS = (
    [os.path.join(os.environ["HOME"], ".config/openstack/"), "/etc/openstack/"]
    if "HOME" in os.environ
    else ["/etc/openstack/"]
)
CONFIG_FILE_NAME = "prometheus.yaml"
LOG = logging.getLogger(__name__)


class ConfigurationError(Exception):
    pass


def get_config_file():
    if os.path.exists(CONFIG_FILE_NAME):
        LOG.debug("Using %s as prometheus configuration", CONFIG_FILE_NAME)
        return open(CONFIG_FILE_NAME)
    for path in DEFAULT_CONFIG_LOCATIONS:
        full_filename = path + CONFIG_FILE_NAME
        if os.path.exists(full_filename):
            LOG.debug("Using %s as prometheus configuration", full_filename)
            return open(full_filename)
    return None


def get_prom_client_from_keystone(session, adapter_options=None):
    if adapter_options is None:
        adapter_options = {}
    endpoint = adapter.Adapter(
        session=session, **adapter_options
    ).get_endpoint()
    parsed_url = parse.urlparse(endpoint)

    escaped_host = netutils.escape_ipv6(parsed_url.hostname)
    root_path = parsed_url.path.strip('/')
    tls = parsed_url.scheme == "https"

    if parsed_url.port is not None:
        url = f'{escaped_host}:{parsed_url.port}'
    else:
        url = escaped_host

    client = PrometheusAPIClient(url, session, root_path)

    if tls:
        client.set_ca_cert(True)
    return client


def get_prom_client_from_file_or_env():
    host = port = ca_cert = None
    root_path = ''
    conf_file = get_config_file()
    if conf_file is not None:
        conf = yaml.safe_load(conf_file)
        if 'host' in conf:
            host = conf['host']
        if 'port' in conf:
            port = conf['port']
        if 'ca_cert' in conf:
            ca_cert = conf['ca_cert']
        if 'root_path' in conf:
            root_path = conf['root_path']
        conf_file.close()
    if 'PROMETHEUS_HOST' in os.environ:
        host = os.environ['PROMETHEUS_HOST']
    if 'PROMETHEUS_PORT' in os.environ:
        port = os.environ['PROMETHEUS_PORT']
    if 'PROMETHEUS_CA_CERT' in os.environ:
        ca_cert = os.environ['PROMETHEUS_CA_CERT']
    if 'PROMETHEUS_ROOT_PATH' in os.environ:
        root_path = os.environ['PROMETHEUS_ROOT_PATH']
    if host is None or port is None:
        raise ConfigurationError("Can't find prometheus host and "
                                 "port configuration in config file or "
                                 "environment variables.")
    escaped_host = netutils.escape_ipv6(host)
    client = PrometheusAPIClient(
        f"{escaped_host}:{port}", None, root_path
    )
    if ca_cert is not None:
        client.set_ca_cert(ca_cert)
    return client


def get_prometheus_client(session=None, adapter_options=None):
    if adapter_options is None:
        adapter_options = {}

    keystone_error = None
    if session is not None:
        try:
            return get_prom_client_from_keystone(session, adapter_options)
        except keystone_exception.EndpointNotFound as e:
            # NOTE(jwysogla): Fallback to get the endpoint configuration from
            # the config file and env vars. If that doesn't work, the error
            # message is raised as ConfigurationError later.
            keystone_error = e
            LOG.debug("Aetos endpoint discovery from Keystone failed: %s", e)

    # NOTE(jwysogla): Always fallback to the original method of discovery
    # through config file and env variables for backwards compatibility.
    try:
        return get_prom_client_from_file_or_env()
    except ConfigurationError as e:
        if keystone_error is not None:
            raise ConfigurationError(
                f"Failed to configure Prometheus client. "
                f"Aetos discovery from keystone failed: '{keystone_error}'. "
                f"Prometheus configuration from config file and environment "
                f"variables failed: '{e}'"
            )
        else:
            raise e


def get_client(obj):
    return obj.app.client_manager.observabilityclient


def format_labels(d: dict) -> str:
    def replace_doubled_quotes(string):
        if "''" in string:
            string = string.replace("''", "'")
        if '""' in string:
            string = string.replace('""', '"')
        return string

    ret = ""
    for key, value in d.items():
        ret += f"{key}='{value}', "
    ret = ret[0:-2]
    old = ""
    while ret != old:
        old = ret
        ret = replace_doubled_quotes(ret)
    return ret


def metrics2cols(m):
    # get all label keys
    cols = list(set().union(*(d.labels.keys() for d in m)))
    cols.sort()
    cols.append("value")
    fields = []
    for metric in m:
        row = [""] * len(cols)
        for key, value in metric.labels.items():
            row[cols.index(key)] = value
        row[cols.index("value")] = metric.value
        fields.append(row)
    return cols, fields



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT = os.path.abspath(os.path.join(BASE_DIR, "..", ".."))

sys.path.insert(0, ROOT)
sys.path.insert(0, BASE_DIR)


def gen_ref(ver, title, names):
    refdir = os.path.join(BASE_DIR, "ref")
    pkg = "observabilityclient"
    if ver:
        pkg = f"{pkg}.{ver}"
        refdir = os.path.join(refdir, ver)
    if not os.path.exists(refdir):
        os.makedirs(refdir)
    idxpath = os.path.join(refdir, "index.rst")
    with open(idxpath, "w") as idx:
        idx.write(("%(title)s\n"
                   "%(signs)s\n"
                   "\n"
                   ".. toctree::\n"
                   "   :maxdepth: 1\n"
                   "\n") % {"title": title, "signs": "=" * len(title)})
        for name in names:
            idx.write("   %s\n" % name)
            rstpath = os.path.join(refdir, "%s.rst" % name)
            with open(rstpath, "w") as rst:
                rst.write(("%(title)s\n"
                           "%(signs)s\n"
                           "\n"
                           ".. automodule:: %(pkg)s.%(name)s\n"
                           "   :members:\n"
                           "   :undoc-members:\n"
                           "   :show-inheritance:\n"
                           "   :noindex:\n")
                          % {"title": " ".join([n.capitalize()
                                                for n in name.split("_")]),
                             "signs": "=" * len(name),
                             "pkg": pkg, "name": name})


gen_ref("v1", "Version 1 API", ["client", "python_api"])
gen_ref("", "rbac", ["rbac"])

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'openstackdocstheme',
    # 'sphinx.ext.intersphinx',
]

# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/python-observabilityclient'
openstackdocs_bug_project = 'python-observabilityclient'
openstackdocs_bug_tag = ''
copyright = '2025, OpenStack Foundation'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
# html_theme = '_theme'
# html_static_path = ['static']
html_theme = 'openstackdocs'

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = ['_theme']

# Output file base name for HTML help builder.
htmlhelp_basename = 'observabilityclientdoc'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ('index',
     'observabilityclient.tex',
     'observabilityclient Documentation',
     'OpenStack Foundation', 'manual'),
]

# Example configuration for intersphinx: refer to the Python standard library.
# intersphinx_mapping = {'http://docs.python.org/': None}
